---
title: Incremental Pre-aggregations
permalink: /incremental-pre-aggregations
scope: cubejs
category: Guides
subCategory: Tutorials
menuOrder: 25
---

When you use partitioned rollups on immutable data there's an opportunity to build pre-aggregations incrementally.
Cube.js allows to set `refreshKey` in a way that partitions in past would be never refreshed without a need which leads to significant performance boost.

As usually there're multiple cubes that require incremental partitions building it's best practice to introduce separate `RefreshKeyHelper.js` file which will contain this reusable logic.
Typical immutable partition rollup will look like:

```javascript
const defaultRefreshKey = `DATE(CURRENT_TIMESTAMP())`;

export const refreshKeySql = `select ${defaultRefreshKey}`;

export const immutablePartitionRollupRefreshKey = (filterFn) =>
  `SELECT CASE WHEN ${filterFn(
    (from, to) => `CURRENT_TIMESTAMP() < TIMESTAMP_ADD(TIMESTAMP(${to}), INTERVAL 72 HOUR)`
  )} THEN ${defaultRefreshKey} END`;
```

Here we use Standard BigQuery SQL dialect.
To use this `refreshKey` with for example `Events` table you can define `autoRollup` like this:

```javascript
import { immutablePartitionRollupRefreshKey } from './RefreshKeyHelper';

cube(`Events`, {
  // ...

  preAggregations: {
    auto: {
      type: `autoRollup`,
      partitionGranularity: `month`,
      refreshKey: {
        sql: immutablePartitionRollupRefreshKey(
          FILTER_PARAMS.FirebaseEvents.time.filter
        )
      }
    },
  }
});
```

In this case SQL generated by `immutablePartitionRollupRefreshKey` returns current date in case `to` date of partition is within 72 hours of current timestamp.
Otherwise it returns `NULL`.
This behavior leads to partitions with `NULL` keys would never refresh until it's SQL is changed.
